@@ -133,8 +133,8 @@ logentry(char *host, char *port, char *qry, char *status)
 
 void
 handlerequest(int sock, char *req, int rlen, char *base, char *ohost,
-	      char *port, char *clienth, char *clientp, int nocgi,
-	      int istls)
+	      char *port, char *clienth, char *clientp, char *serverh,
+	      char *serverp, int nocgi, int istls)
 {
 	struct stat dir;
 	char recvc[1025], recvb[1025], path[1025], args[1025], argsc[1025],
@@ -367,7 +367,7 @@ handlerequest(int sock, char *req, int rlen, char *base, char *ohost,
 				logentry(clienth, clientp, recvc, "nocgi error");
 		} else {
 			type->f(sock, path, port, base, args, sear, ohost,
-				clienth, istls);
+				clienth, serverh, istls);
 		}
 	} else {
 		/*
@@ -376,7 +376,7 @@ handlerequest(int sock, char *req, int rlen, char *base, char *ohost,
 		 */
 		if (!pathfallthrough && S_ISDIR(dir.st_mode)) {
 			handledir(sock, path, port, base, args, sear, ohost,
-				clienth, istls);
+				clienth, serverh, istls);
 			if (loglvl & DIRS) {
 				logentry(clienth, clientp, recvc,
 							"dir listing");
@@ -526,7 +526,7 @@ getlistenfd(struct addrinfo *hints, char *bindip, char *port, int *rlfdnum)
 void
 usage(void)
 {
-	dprintf(2, "usage: %s [-46cdens] [-l logfile] "
+	dprintf(2, "usage: %s [-46cdensy] [-l logfile] "
 #ifdef ENABLE_TLS
 		   "[-t keyfile certfile] "
 #endif /* ENABLE_TLS */
@@ -540,13 +540,13 @@ int
 main(int argc, char *argv[])
 {
 	struct addrinfo hints;
-	struct sockaddr_storage clt;
+	struct sockaddr_storage clt, slt;
 	struct linger lingerie;
-	socklen_t cltlen;
+	socklen_t cltlen, sltlen;
 	int sock, dofork = 1, inetf = AF_UNSPEC, usechroot = 0,
 	    nocgi = 0, errno_save, nbindips = 0, i, j,
 	    nlfdret, *lfdret, listfd, maxlfd, istls = 0,
-	    dotls = 0,
+	    dotls = 0, dohaproxy = 0, tcpver = -1, haret = 0,
 #ifdef ENABLE_TLS
 	    tlspipe[2], shufbuf[1025],
 	    shuflen, wlen, shufpos,
@@ -556,11 +556,13 @@ main(int argc, char *argv[])
 	fd_set rfd;
 	char *port, *base, clienth[NI_MAXHOST], clientp[NI_MAXSERV],
 	     *user = NULL, *group = NULL, **bindips = NULL,
-	     *ohost = NULL, *sport = NULL, *p,
+	     *ohost = NULL, *sport = NULL, *p;
+	/* Must be as large as recvb, due to scanf restrictions. */
+	char hachost[1025], hashost[1025], hacport[1025], hasport[1025],
 #ifdef ENABLE_TLS
 	     *certfile = NULL, *keyfile = NULL,
 #endif /* ENABLE_TLS */
-	     byte0, recvb[1025];
+	     byte0, recvb[1025], serverh[NI_MAXHOST], serverp[NI_MAXSERV];
 	struct passwd *us = NULL;
 	struct group *gr = NULL;
 #ifdef ENABLE_TLS
@@ -574,9 +576,11 @@ main(int argc, char *argv[])
 	ARGBEGIN {
 	case '4':
 		inetf = AF_INET;
+		tcpver = 4;
 		break;
 	case '6':
 		inetf = AF_INET6;
+		tcpver = 6;
 		break;
 	case 'b':
 		base = EARGF(usage());
@@ -630,6 +634,9 @@ main(int argc, char *argv[])
 	case 'v':
 		loglvl = atoi(EARGF(usage()));
 		break;
+	case 'y':
+		dohaproxy = 1;
+		break;
 	default:
 		usage();
 	} ARGEND;
@@ -808,6 +815,10 @@ main(int argc, char *argv[])
 
 	initsignals();
 
+#ifdef HOT_COMPUTER
+#warning "I love you too."
+#endif
+
 #ifdef __OpenBSD__
 	char promises[31]; /* check the size needed in the fork too */
 	snprintf(promises, sizeof(promises), "rpath inet stdio proc exec %s",
@@ -858,6 +869,16 @@ main(int argc, char *argv[])
 			}
 		}
 
+		sltlen = sizeof(slt);
+		serverh[0] = serverp[0] = '\0';
+		if (getsockname(sock, (struct sockaddr *)&slt, &sltlen) == 0) {
+			getnameinfo((struct sockaddr *)&slt, sltlen, serverh,
+					sizeof(serverh), serverp, sizeof(serverp),
+					NI_NUMERICHOST|NI_NUMERICSERV);
+		}
+		if (!strncmp(serverh, "::ffff:", 7))
+			memmove(serverh, serverh+7, strlen(serverh)-6);
+
 		if (getnameinfo((struct sockaddr *)&clt, cltlen, clienth,
 				sizeof(clienth), clientp, sizeof(clientp),
 				NI_NUMERICHOST|NI_NUMERICSERV)) {
@@ -895,6 +916,10 @@ main(int argc, char *argv[])
 			}
 #endif /* __OpenBSD__ */
 
+read_selector_again:
+			rlen = 0;
+			memset(recvb, 0, sizeof(recvb));
+
 			if (recv(sock, &byte0, 1, MSG_PEEK) < 1)
 				return 1;
 
@@ -933,14 +958,14 @@ main(int argc, char *argv[])
 #ifdef ENABLE_TLS
 				if (istls) {
 					retl = tls_read(tlsclientctx,
-						recvb+rlen, sizeof(recvb)-1-rlen);
+						recvb+rlen, 1);
 					if (retl < 0)
 						fprintf(stderr, "tls_read failed: %s\n", tls_error(tlsclientctx));
 				} else
 #endif /* ENABLE_TLS */
 				{
 					retl = read(sock, recvb+rlen,
-						sizeof(recvb)-1-rlen);
+						1);
 					if (retl < 0)
 						perror("read");
 				}
@@ -952,6 +977,59 @@ main(int argc, char *argv[])
 			if (rlen <= 0)
 				return 1;
 
+			/*
+			 * HAProxy v1 protocol support.
+			 * TODO: Add other protocol version support.
+			 */
+			if (dohaproxy && !strncmp(recvb, "PROXY TCP", 9)) {
+				if (p[-1] == '\r')
+					p[-1] = '\0';
+				*p++ = '\0';
+
+				/*
+				 * Be careful, we are using scanf.
+				 * TODO: Use some better parsing.
+				 */
+				memset(hachost, 0, sizeof(hachost));
+				memset(hashost, 0, sizeof(hashost));
+				memset(hacport, 0, sizeof(hacport));
+				memset(hasport, 0, sizeof(hasport));
+
+				haret = sscanf(recvb, "PROXY TCP%d %s %s %s %s",
+					&tcpver, hachost, hashost, hacport,
+					hasport);
+				if (haret != 5)
+					return 1;
+
+				/*
+				 * Be careful. Everything could be
+				 * malicious.
+				 */
+				memset(clienth, 0, sizeof(clienth));
+				memmove(clienth, hachost, sizeof(clienth)-1);
+				memset(serverh, 0, sizeof(serverh));
+				memmove(serverh, hashost, sizeof(serverh)-1);
+				memset(clientp, 0, sizeof(clientp));
+				memmove(clientp, hacport, sizeof(clientp)-1);
+				memset(serverp, 0, sizeof(serverp));
+				memmove(serverp, hasport, sizeof(serverp)-1);
+
+				if (!strncmp(serverh, "::ffff:", 7)) {
+					memmove(serverh, serverh+7,
+							strlen(serverh)-6);
+				}
+				if (!strncmp(clienth, "::ffff:", 7)) {
+					memmove(clienth, clienth+7,
+							strlen(clienth)-6);
+				}
+				if (loglvl & CONN) {
+					logentry(clienth, clientp, "-",
+							"haproxy connection");
+				}
+
+				goto read_selector_again;
+			}
+
 #ifdef ENABLE_TLS
 			if (istls) {
 				if (pipe(tlspipe) < 0) {
@@ -995,8 +1073,10 @@ main(int argc, char *argv[])
 #endif /* ENABLE_TLS */
 
 			handlerequest(sock, recvb, rlen, base,
-					ohost, sport, clienth,
-					clientp, nocgi, istls);
+					(dohaproxy)? serverh : ohost,
+					(dohaproxy)? serverp : sport,
+					clienth, clientp, serverh, serverp,
+					nocgi, istls);
 
 			if (!istls) {
 				/*
